package org.roboid.robot.impl;

import org.roboid.robot.MotoringDevice;

/**
 * @author akaii@kw.ac.kr (Kwang-Hyun Park)
 */
abstract class IntMotoringDeviceImpl extends IntDeviceImpl implements MotoringDevice
{
	private final Object mWriteLock;
	private int[] mWriteData;
	
	IntMotoringDeviceImpl(RoboidImpl parent, int uid, String name, int dataSize, Object initialValue, Object readLock, Object writeLock, Object fireLock)
	{
		super(parent, uid, name, dataSize, initialValue, readLock, fireLock);
		
		mWriteLock = writeLock;
		if(dataSize < 0) return;
		
		mWriteData = new int[dataSize];
		fill(mWriteData, initialValue, 0, dataSize);
	}
	
	@Override
	Object getWriteData()
	{
		return mWriteData;
	}

	@Override
	public boolean write(int data)
	{
		synchronized(mWriteLock)
		{
			if(mWriteData == null || mWriteData.length <= 0)
			{
				if(mDataSize < 0)
					mWriteData = new int[1];
				else
					return false;
			}
			mWriteData[0] = data;
			fire();
			notifyMotoringDeviceWritten();
		}
		return true;
	}

	@Override
	public boolean write(int index, int data)
	{
		if(index < 0) return false;
		synchronized(mWriteLock)
		{
			if(mWriteData == null)
			{
				if(mDataSize < 0)
				{
					mWriteData = new int[index + 1];
					fill(mWriteData, mInitialValue, 0, index);
				}
				else
					return false;
			}
			else if(index >= mWriteData.length)
			{
				if(mDataSize < 0)
				{
					int[] newData = new int[index + 1];
					int len = mWriteData.length;
					System.arraycopy(mWriteData, 0, newData, 0, len);
					fill(newData, mInitialValue, len, index);
					mWriteData = newData;
				}
				else
					return false;
			}
			mWriteData[index] = data;
			fire();
			notifyMotoringDeviceWritten();
		}
		return true;
	}

	@Override
	public int write(int[] data)
	{
		if(data == null) return 0;
		int datalen = data.length;
		if(datalen <= 0) return 0;
		
		int len = 0;
		synchronized(mWriteLock)
		{
			if(mDataSize < 0)
			{
				if(mWriteData == null || mWriteData.length != datalen)
					mWriteData = new int[datalen];
			}
			int[] writeData = mWriteData;
			if(writeData == null) return 0;
			int thislen = writeData.length;
			if(thislen <= 0) return 0;
			len = Math.min(thislen, datalen);
			System.arraycopy(data, 0, writeData, 0, len);
			for(int i = len; i < thislen; ++i)
				writeData[i] = 0;
			fire();
			notifyMotoringDeviceWritten();
		}
		return len;
	}

	@Override
	public boolean writeFloat(float data)
	{
		return write((int)data);
	}

	@Override
	public boolean writeFloat(int index, float data)
	{
		return write(index, (int)data);
	}

	@Override
	public int writeFloat(float[] data)
	{
		if(data == null) return 0;
		int datalen = data.length;
		if(datalen <= 0) return 0;
		
		int len = 0;
		synchronized(mWriteLock)
		{
			if(mDataSize < 0)
			{
				if(mWriteData == null || mWriteData.length != datalen)
					mWriteData = new int[datalen];
			}
			int[] writeData = mWriteData;
			if(writeData == null) return 0;
			int thislen = writeData.length;
			if(thislen <= 0) return 0;
			len = Math.min(thislen, datalen);
			for(int i = 0; i < len; ++i)
				writeData[i] = (int)data[i];
			for(int i = len; i < thislen; ++i)
				writeData[i] = 0;
			fire();
			notifyMotoringDeviceWritten();
		}
		return len;
	}
}
