package org.roboid.robot.impl;

import java.util.ArrayList;

import org.roboid.robot.DataType;
import org.roboid.robot.Device;
import org.roboid.robot.DeviceDataChangedListener;
import org.roboid.robot.DeviceType;
import org.roboid.robot.Roboid;

/**
 * @author akaii@kw.ac.kr (Kwang-Hyun Park)
 */
public abstract class RoboidImpl extends NamedElementImpl implements Roboid
{
	private int mTag;
	private final ArrayList<RoboidImpl> mRoboids = new ArrayList<RoboidImpl>();
	protected final ArrayList<DeviceDataChangedListener> mDeviceDataChangedListeners = new ArrayList<DeviceDataChangedListener>();
	protected final DeviceImpl[] mDevices;
	protected final Object mReadLock = new Object();
	protected final Object mWriteLock = new Object();
	private final Object mFireLock = new Object();

	protected RoboidImpl(int size)
	{
		mDevices = new DeviceImpl[size];
	}
	
	protected RoboidImpl(int size, int tag)
	{
		mDevices = new DeviceImpl[size];
		mTag = tag;
	}
	
	protected void setTag(int tag)
	{
		mTag = tag;
	}
	
	protected DeviceImpl addDevice(int index, int deviceType, int uid, String name, int dataType, int dataSize, Object initialValue)
	{
		if(index < 0 || index >= mDevices.length) return null;
		
		DeviceImpl device = null;
		switch(deviceType)
		{
		case DeviceType.SENSOR:
			switch(dataType)
			{
			case DataType.INTEGER:
				device = new IntSensorImpl(this, uid, name, dataSize, initialValue, mReadLock, mFireLock);
				break;
			case DataType.FLOAT:
				device = new FloatSensorImpl(this, uid, name, dataSize, initialValue, mReadLock, mFireLock);
				break;
			case DataType.STRING:
				device = new StringSensorImpl(this, uid, name, dataSize, initialValue, mReadLock, mFireLock);
				break;
			}
			break;
		case DeviceType.EFFECTOR:
			switch(dataType)
			{
			case DataType.INTEGER:
				device = new IntEffectorImpl(this, uid, name, dataSize, initialValue, mReadLock, mWriteLock, mFireLock);
				break;
			case DataType.FLOAT:
				device = new FloatEffectorImpl(this, uid, name, dataSize, initialValue, mReadLock, mWriteLock, mFireLock);
				break;
			case DataType.STRING:
				device = new StringEffectorImpl(this, uid, name, dataSize, initialValue, mReadLock, mWriteLock, mFireLock);
				break;
			}
			break;
		case DeviceType.EVENT:
			switch(dataType)
			{
			case DataType.INTEGER:
				device = new IntEventImpl(this, uid, name, dataSize, initialValue, mReadLock, mFireLock);
				break;
			case DataType.FLOAT:
				device = new FloatEventImpl(this, uid, name, dataSize, initialValue, mReadLock, mFireLock);
				break;
			case DataType.STRING:
				device = new StringEventImpl(this, uid, name, dataSize, initialValue, mReadLock, mFireLock);
				break;
			}
			break;
		case DeviceType.COMMAND:
			switch(dataType)
			{
			case DataType.INTEGER:
				device = new IntCommandImpl(this, uid, name, dataSize, initialValue, mReadLock, mWriteLock, mFireLock);
				break;
			case DataType.FLOAT:
				device = new FloatCommandImpl(this, uid, name, dataSize, initialValue, mReadLock, mWriteLock, mFireLock);
				break;
			case DataType.STRING:
				device = new StringCommandImpl(this, uid, name, dataSize, initialValue, mReadLock, mWriteLock, mFireLock);
				break;
			}
			break;
		}
		if(device != null)
			mDevices[index] = device;
		return device;
	}
	
	protected void addRoboid(RoboidImpl roboid)
	{
		if(roboid == null) return;
		synchronized(mRoboids)
		{
			if(mRoboids.contains(roboid)) return;
			mRoboids.add(roboid);
		}
		synchronized(mDeviceDataChangedListeners)
		{
			for(DeviceDataChangedListener listener : mDeviceDataChangedListeners)
				roboid.addDeviceDataChangedListener(listener);
		}
	}
	
	protected Object getReadData(Device device)
	{
		if(device == null) return null;
		return ((DeviceImpl)device).getReadData();
	}
	
	protected Object getWriteData(Device device)
	{
		if(device == null) return null;
		return ((DeviceImpl)device).getWriteData();
	}
	
	protected boolean putReadData(DeviceImpl device, Object data)
	{
		if(device == null) return false;
		return device.putReadData(data);
	}
	
	@Override
	public Device findDeviceById(int deviceId)
	{
		return findDevice(deviceId);
	}
	
	protected Device findDevice(int uid)
	{
		int tag = uid & 0xfffff000;
		if(tag == mTag)
		{
			for(DeviceImpl device : mDevices)
			{
				if(device.getUid() == uid)
					return device;
			}
			return null;
		}
		synchronized(mRoboids)
		{
			Device device = null;
			for(RoboidImpl roboid : mRoboids)
			{
				device = roboid.findDevice(uid);
				if(device != null)
					return device;
			}
		}
		return null;
	}
	
	@Override
	public void addDeviceDataChangedListener(DeviceDataChangedListener listener)
	{
		if(listener == null) return;
		synchronized(mDeviceDataChangedListeners)
		{
			if(!mDeviceDataChangedListeners.contains(listener))
				mDeviceDataChangedListeners.add(listener);
		}
		synchronized(mRoboids)
		{
			for(Roboid roboid : mRoboids)
				roboid.addDeviceDataChangedListener(listener);
		}
	}
	
	@Override
	public void removeDeviceDataChangedListener(DeviceDataChangedListener listener)
	{
		if(listener == null) return;
		synchronized(mDeviceDataChangedListeners)
		{
			mDeviceDataChangedListeners.remove(listener);
		}
		synchronized(mRoboids)
		{
			for(Roboid roboid : mRoboids)
				roboid.removeDeviceDataChangedListener(listener);
		}
	}
	
	@Override
	public void clearDeviceDataChangedListeners()
	{
		synchronized(mDeviceDataChangedListeners)
		{
			mDeviceDataChangedListeners.clear();
		}
		synchronized(mRoboids)
		{
			for(Roboid roboid : mRoboids)
				roboid.clearDeviceDataChangedListeners();
		}
	}
	
	protected void fire(int index)
	{
		if(index < 0 || index >= mDevices.length) return;
		mDevices[index].fire();
	}
	
	protected void onMotoringDeviceWritten(Device device)
	{
	}
	
	protected void updateDeviceState()
	{
		for(DeviceImpl device : mDevices)
			device.updateDeviceState();
		synchronized(mRoboids)
		{
			for(RoboidImpl roboid : mRoboids)
				roboid.updateDeviceState();
		}
	}
}
